#ifdef WEBGL
#ifdef MULTIDRAW
#extension GL_ANGLE_multi_draw : require
#endif
#endif

precision highp float;

/* Normal and tangent are float16, but we transform them with highp quats
 * anyway so do the conversion as early as possible */
layout(location = 0) in highp vec3 inPosition;
#ifdef TEXTURE_COORDS
layout(location = 1) in highp vec2 inTextureCoords;
#endif
#ifdef COLOR
layout(location = 2) in mediump vec4 inColor;
#endif
#ifdef TANGENT
layout(location = 3) in highp vec4 inTangent;
#endif
#ifdef BATCH
#ifndef MULTIDRAW
layout(location = 4) in mediump uint inObjectId;
#endif
#endif
#ifdef NORMAL
layout(location = 5) in highp vec3 inNormal;
#endif

#ifdef QUATERNION
    struct Quat2 {
        vec4 rot;
        vec4 loc;
    };

    vec4 conjugated(vec4 q) {
        return vec4(-q.xyz, q.w);
    }

    Quat2 conjugated(Quat2 q) {
        return Quat2(conjugated(q.rot), vec4(q.loc.xyz, -q.loc.w));
    }

    vec4 quat_mult(vec4 q1, vec4 q2) {
        return vec4(q1.w*q2.xyz + q2.w*q1.xyz + cross(q1.xyz, q2.xyz),
                    //dot(q1, conjugated(q2)));
                    q1.w*q2.w - dot(q1.xyz, q2.xyz));
    }

    Quat2 quat2_mult(vec3 p, Quat2 q) {
        return Quat2(q.rot, q.loc + quat_mult(vec4(p, 0.0f), q.rot));
    }

    Quat2 quat2_mult(Quat2 a, Quat2 b) {
        return Quat2(quat_mult(a.rot, b.rot), quat_mult(a.rot, b.loc) + quat_mult(a.loc, b.rot));
    }

    vec3 quat_transformVector(vec4 q, vec3 v) {
        vec3 a = cross(q.xyz, v);
        vec3 b = q.w*a + cross(q.xyz, a);
        return v + 2.0*b;
    }

    vec3 quat2_transformPoint(Quat2 q, vec3 p) {
        vec3 qrv = q.rot.xyz;
        float qrw = q.rot.w;

        vec3 blv = qrw*p + cross(qrv, p) + 2.0*q.loc.xyz;
        return (dot(p, qrv) - 2.0*q.loc.w)*qrv
            + qrw*blv
            + cross(qrv, blv);
    }
#endif

#ifdef POSITION_WORLD
out highp vec3 fragPositionWorld;
#endif
#ifdef POSITION_VIEW
out highp vec3 fragPositionView;
#endif
#ifdef TEXTURE_COORDS
out highp vec2 fragTextureCoords;
#endif
#ifdef COLOR
out mediump vec4 fragColor;
#endif
#ifdef TANGENT
out mediump vec4 fragTangent;
#endif
#ifdef OBJECT_ID
flat out mediump uint fragObjectId;
#endif
#ifdef NORMAL
out mediump vec3 fragNormal;
#endif
#ifdef BARYCENTRIC
out mediump vec3 fragBarycentric;
#endif

uniform highp mat4 projectionMatrix;
#ifndef MULTIDRAW
uniform mediump uint idOffset;
#endif

#ifdef QUATERNION
    #ifdef BATCH
    uniform Transformations {
        vec4 transformations[NUM_OBJECTS*2u];
        vec4 scalings[NUM_OBJECTS];
    };
    #else
    uniform highp vec4 transform[2];
    uniform highp vec3 scaling;
    #endif

    uniform highp vec4 viewTransform[2];
#else
    #ifdef BATCH
    uniform Transformations {
        mat4 transformations[NUM_OBJECTS];
    };
    #else
    uniform highp mat4 transform;
    #endif

    uniform highp mat4 viewTransform;
    uniform highp mat3 normalMat;
#endif

void main() {
/* Temporaries since we have no out variables */
#ifndef POSITION_WORLD
highp vec3 fragPositionWorld;
#endif
#ifndef POSITION_VIEW
highp vec3 fragPositionView;
#endif
#ifndef OBJECT_ID
mediump uint fragObjectId;
#endif

#ifdef QUATERNION
    #ifdef BATCH
        #ifdef MULTIDRAW
        fragObjectId = uint(gl_DrawID); /* idOffset not needed! */
        #else
        fragObjectId = inObjectId - idOffset;
        #endif
        highp vec4 transform[2] = vec4[](transformations[2u*fragObjectId], transformations[2u*fragObjectId+1u]);
        highp vec3 scaling = scalings[fragObjectId].xyz;
    #else
        fragObjectId = 0u;
    #endif
    /* Transformed vertex position */
    fragPositionWorld = quat2_transformPoint(Quat2(transform[0], transform[1]), scaling.xyz*inPosition);
    fragPositionView = quat2_transformPoint(Quat2(viewTransform[0], viewTransform[1]), fragPositionWorld);

    /* Transformed normal vector */
    #ifdef NORMAL
    fragNormal = normalize(quat_transformVector(transform[0], scaling*inNormal));
    #endif
    #ifdef TANGENT
    fragTangent = vec4(normalize(quat_transformVector(transform[0], scaling*inTangent.xyz)), inTangent.w);
    #endif
#else
    #ifdef BATCH
        #ifdef MULTIDRAW
        fragObjectId = uint(gl_DrawID) - idOffset;
        #else
        fragObjectId = inObjectId - idOffset;
        #endif
        mat4 transform = transformations[fragObjectId];
    #else
        fragObjectId = 0u;
    #endif
    /* Transformed vertex position */
    fragPositionWorld = (transform*vec4(inPosition, 1.0)).xyz;
    fragPositionView = (viewTransform*vec4(fragPositionWorld, 1.0)).xyz;

    /* Transformed normal vector */
    #ifdef NORMAL
    fragNormal = normalMat*inNormal;
    #endif
    #ifdef TANGENT
    fragTangent = normalMat*inTangent;
    #endif
#endif

    /* Transform the position */
    gl_Position = projectionMatrix*vec4(fragPositionView, 1.0f);

#ifdef TEXTURE_COORDS
    /* Texture coordinates, if needed */
    fragTextureCoords = inTextureCoords;
#endif

#ifdef COLOR
    fragColor = inColor;
#endif

#ifdef BARYCENTRIC
    fragBarycentric = vec3(0.0);
    #ifndef NEW_GLSL
    fragBarycentric[gl_VertexID % 3] = 1.0;
    #else
    fragBarycentric[int(mod(float(gl_VertexID), 3.0))] = 1.0;
    #endif
#endif
}
