#ifdef WEBGL
#ifdef MULTIDRAW
#extension GL_ANGLE_multi_draw : require
#endif
#endif

precision highp float;
precision mediump int;

/* Normal and tangent are float16, but we transform them with highp quats
 * anyway so do the conversion as early as possible */
layout(location = 0) in highp vec3 inPosition;
#ifdef TEXTURE_COORDS
layout(location = 1) in highp vec2 inTextureCoords;
#endif
#ifdef COLOR
layout(location = 2) in mediump vec4 inColor;
#endif
#ifdef TANGENT
layout(location = 3) in highp vec4 inTangent;
#endif
#ifndef MULTIDRAW
layout(location = 4) in mediump uint inObjectId;
#endif
#ifdef NORMAL
layout(location = 5) in highp vec3 inNormal;
#endif

layout(location = 6) in mediump uvec4 jointIds;
layout(location = 7) in mediump vec4 weights;
#ifdef SECONDARY_JOINTS
layout(location = 8) in mediump uvec4 secondaryJointIds;
layout(location = 9) in mediump vec4 secondaryWeights;
#endif

#include "lib/Quaternion.glsl"
#include "lib/Packing.frag"

/* Order should match the location ids above. Precisions match the underlying
 * formats to avoid a conversion.
 * The uvec2 outputs are normalized short vec3 and vec4 that we manually pack
 * because transform feedback can only write out full floats and uints. */
out highp vec3 positionWorld;
#ifdef TEXTURE_COORDS
out highp vec2 textureCoords;
#endif
#ifdef COLOR
flat out highp uvec2 color;
#endif
#ifdef TANGENT
flat out highp uvec2 tangent;
#endif
#ifndef MULTIDRAW
flat out highp uint objectId;
#endif
#ifdef NORMAL
flat out highp uvec2 normal;
#endif

uniform highp sampler2DArray skinDataTexture;
#ifndef MULTIDRAW
uniform mediump uint idOffset;
#endif
/* ushort skin IDs are packed into uints */
uniform highp uint skinList[SKIN_COUNT/2u];

Quat2 jointTransform(uint layer, uint row, uint joint) {
    ivec3 coord = ivec3(3*int(joint), int(row), int(layer));
    return Quat2(
        texelFetch(skinDataTexture, coord, 0),
        texelFetchOffset(skinDataTexture, coord, 0, ivec2(1, 0)));
}

vec3 jointScaling(uint layer, uint row, uint joint) {
    ivec3 coord = ivec3(3*int(joint), int(row), int(layer));
    return texelFetchOffset(skinDataTexture, coord, 0, ivec2(2, 0)).xyz;
}

float copysign(float from, float to) {
    const uint firstBit = (1u << 31u);
    const uint otherBits = ~(1u << 31u);
    return uintBitsToFloat(
        /* Multiply the signs only */
        ((floatBitsToUint(from) ^ floatBitsToUint(to)) & firstBit) |
        /* Keep non-sign bits the same */
        (floatBitsToUint(to) & otherBits));
}

void main() {
    #ifdef MULTIDRAW
    uint drawId = uint(gl_DrawID);
    #else
    uint drawId = inObjectId - idOffset;
    #endif

    /* Two ushort skin IDs are packed into the lower/upper word of one uint */
    uint skin = (skinList[drawId/2u] >> ((drawId & 1u) * 16u)) & 0xffffu;

    uint layer = skin >> NUM_ROWS_LOG2;
    uint row = skin & ((1u << NUM_ROWS_LOG2) - 1u);

    Quat2 jt[] = Quat2[](
        jointTransform(layer, row, jointIds[0]),
        jointTransform(layer, row, jointIds[1]),
        jointTransform(layer, row, jointIds[2]),
        jointTransform(layer, row, jointIds[3])
    #ifdef SECONDARY_JOINTS
        ,
        jointTransform(layer, row, secondaryJointIds[0]),
        jointTransform(layer, row, secondaryJointIds[1]),
        jointTransform(layer, row, secondaryJointIds[2]),
        jointTransform(layer, row, secondaryJointIds[3])
    #endif
    );

    vec3 js[] = vec3[](
        jointScaling(layer, row, jointIds[0]),
        jointScaling(layer, row, jointIds[1]),
        jointScaling(layer, row, jointIds[2]),
        jointScaling(layer, row, jointIds[3])
    #ifdef SECONDARY_JOINTS
        ,
        jointScaling(layer, row, secondaryJointIds[0]),
        jointScaling(layer, row, secondaryJointIds[1]),
        jointScaling(layer, row, secondaryJointIds[2]),
        jointScaling(layer, row, secondaryJointIds[3])

    #endif
    );

    float w[] = float[](
        weights[0],
        weights[1],
        weights[2],
        weights[3]
    #ifdef SECONDARY_JOINTS
        ,
        secondaryWeights[0],
        secondaryWeights[1],
        secondaryWeights[2],
        secondaryWeights[3]
    #endif
    );

    const int jointsPerVertex =
        #ifdef SECONDARY_JOINTS
        8;
        #else
        4;
        #endif

    Quat2 skinTransform = quat_scale(w[0], jt[0]);
    for(int i = 1; i != jointsPerVertex; ++i) {
        /* Handle antipodality */
        float adjustedW = copysign(dot(jt[0].rot, jt[i].rot), w[i]);
        Quat2 t = quat_scale(adjustedW, jt[i]);
        skinTransform.rot += t.rot;
        skinTransform.loc += t.loc;
    }

    float len = 1.0/length(skinTransform.rot);
    skinTransform.rot *= len;
    skinTransform.loc *= len;

    vec3 scaling = w[0]*js[0];
    float totalW = w[0];
    for(int i = 1; i != jointsPerVertex; ++i) {
        scaling += w[i]*js[i];
        totalW += w[i];
    }
    scaling /= totalW;

    positionWorld = quat2_transformPoint(skinTransform, scaling*inPosition);
    #ifdef TANGENT
    tangent = packSnormVector(vec4(quat_transformVector(skinTransform.rot, inTangent.xyz), inTangent.w));
    #endif
    #ifdef NORMAL
    normal = packSnormVector(vec4(quat_transformVector(skinTransform.rot, inNormal), 0.0));
    #endif

    /* Passthrough attributes */
    #ifdef TEXTURE_COORDS
    textureCoords = inTextureCoords;
    #endif
    #ifdef COLOR
    color = packUnormVector(inColor);
    #endif
    #ifndef MULTIDRAW
    objectId = inObjectId;
    #endif
}
